#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time
import logging
from threading import Event, Thread
from math import ceil
import Queue

from PowerUnit import PowerUnit
from Pid import PID
from RateLimiter import RateLimiter
from AdaptiveStromregelung import AdaptiveStromregelung

_logger = logging.getLogger(__name__)
ENABLE_DEBUG_LOG = False

MIN_TARGET_TEMPERATURE = 298    # 298 K oder 25°C
RATING_INTERVAL = 1             # Ratelimiter wird jede Sekunde ausgeführt
F700_TEMP_DEVIATION = 10        # Temperaturabweichung [K], ab der Fehler F700 aktiv wird
F70_TEMP_DEVIATION = 30         # Temperaturabweichung [K], ab der Fehler F70 aktiv wird
F701_TIMEOUT = 10               # Timeout [min], ab dem Fehler F701 aktiv wird
MIN_HEATING_GRADIENT = 0.1      # Minimal geforderte Heizrate K/min


class HeatFileProcessing(object):
    """!
    Diese Klasse ist für die Ablaufsteuerung des Heizprozesses verantwortlich.
    Hinweis: Alle Temperaturwerte in [K].
    """
    def __init__(self, gs1controller, config_manager, doorcontroller, analog_input, error_handler):
        """!
        Initialisiert die Abarbeitung einer Heizdatei.
        @param gs1controller: Instanz des Gs1Controllers
        @param config_manager: Instanz des Konfig Managers
        @param doorcontroller: Instanz der Türsteuerung
        @param analog_input: Instanz des AnalogInputs
        @param error_handler: Instanz des Error Handlers
        """
        _logger.debug("HeatFileProcessing init")
        self._name = "HeatFileProcessing"

        self._gs1Controller = gs1controller
        self._rate_limiter = None
        self._analog_input = analog_input
        self._error_handler = error_handler
        self._doorController = doorcontroller
        self._pid = PID(config_manager)
        self._pu = PowerUnit(config_manager, error_handler)
        self._output_buffer = Queue.Queue()

        self._hfStages = []                                                 # Liste der Stufen für den Heizprozess
        self._state = "IDLE"                                                # Zustand Heizprozess
        self._heatingProcessState = ['0', '0', 'hold']                      # Zustandsinfo für GUI
        self._heatingResult = True                                          # Heizprogramm (nicht) erfolgreich
        self._activeStage = 0                                               # Laufende Nummer der aktiven Stufe
        self._heat_rate = 0                                                 # Heizrate
        self._hold_time = 0                                                 # Haltezeit
        self._actuator_set_point = 0                                        # Steuerwert Leistungseinheit
        self._startHeatingProcess = False                                   # True: Heizvorgang starten
        self._preDryDuration = 0                                            # Dauer der Vortrockungsphase
        self._userInteraction = False                                       # True: Benutzerinteraktion ist erfolgt
        self._actual_heating_time = 0                                       # Aktuelle Gesamtdauer des laufenden Heizprozesses
        self._estimated_heating_time = 0                                    # Geschätzte Restlaufzeit des laufenden Heizprozesses
        self._last_target_temperature = MIN_TARGET_TEMPERATURE              # Zieltemperatur der vergangenen Stufe. Wird auch von Rampengenerator verwendet!
        self._target_temperature = MIN_TARGET_TEMPERATURE                   # Zieltemperatur der ausgeführten Stufe
        self._min_heating_gradient = MIN_HEATING_GRADIENT * F701_TIMEOUT    # tatsächlichen Gradient auf Grund von Überwachungsintervall berechnen
        self._f701_timeout = F701_TIMEOUT * 60                              # in Sekunden umrechnen
        self._target_temp_timeout = time.time() + self._f701_timeout        # Timer für Fehler F701: Endtemperatur nicht erreichbar
        self._last_F701_temp = MIN_TARGET_TEMPERATURE                       # Merker für Fehlerüberwachung F701

        # Toleranzband für Soll-Temperatur
        self._set_point_tolerance = float(config_manager.get_config_file_value('heater', 'set_point_tolerance'))
        # ; Heizrate beim Vortrocknen [K/min], geteilt durch 60, da in [K/min]
        self._pre_dry_heatingrate = float(config_manager.get_config_file_value('heater', 'pre_dry_heatingrate')) / 60.0
        # Brennkammer Soll-Temperatur beim Vortrocknen
        self._pre_dry_temperature = int(config_manager.get_config_file_value('heater', 'pre_dry_temperature'))
        # Maximal erlaubte Brennraum-Temperatur, wenn die Türe offen ist
        self._max_temp_door_open = float(config_manager.get_config_file_value('door', 'max-temperature'))
        # Maximal erlaubte Brennraum-Temperatur
        self._limit_temperature_cc = float(config_manager.get_config_file_value('temp_limits', 'limit-temperature-cc'))

        # Türkühlung ein / aus
        self._linear_cooling = True if config_manager.get_config_file_value('lin_cooling', 'lin-cooling') == 'yes' else False
        # Schrittweite für Türkühlung in %
        self._door_step_value = int(config_manager.get_config_file_value('lin_cooling', 'step-value-door'))
        # Maximaler Wert, wie weit die Türe beim linearen Abkühlen geöffnet werden darf [%]
        self._linear_cooling_max_value = int(config_manager.get_config_file_value('lin_cooling', 'max-lin-cooling-percent'))
        # Zeit zwischen zwei Türöffnungen bei linearem Abkühlen
        self._door_remain_time_cooling = int(config_manager.get_config_file_value('lin_cooling', 'remain-time-lin-cooling')) / 1000.0  # 60
        # Maximal erlaubte Brennraumtemperatur für lineares Abkühlen
        self._door_max_temp_cooling = int(config_manager.get_config_file_value('lin_cooling', 'max-temp-lin-cooling'))
        # P-Wert, bei welchem lineares Abkühlen beginnt
        self._k_p_lin_cooling = int(config_manager.get_config_file_value('lin_cooling', 'k_p_lin_cooling'))

        # Kennlinie für Adaptive Stromregelung lesen
        t1 = float(config_manager.get_config_file_value('current_limiter', 't1'))
        i1 = float(config_manager.get_config_file_value('current_limiter', 'i1'))
        t2 = float(config_manager.get_config_file_value('current_limiter', 't2'))
        i2 = float(config_manager.get_config_file_value('current_limiter', 'i2'))
        # Adaptive Stromregelung initialisieren
        self._asr = AdaptiveStromregelung(t1, i1, t2, i2)

        self._process_time = 0.2        # 200 msec process to calculate everything

        self._stopped = Event()
        self._stopped.clear()

        self._heating_thread = Thread(target=self._process, name=self._name)
        self._heating_thread.setDaemon(True)
        self._heating_thread.start()

    def exit(self):
        """!
        Beendet Thread und löscht selbst erzeugte Objekte
        """
        self._stopped.set()
        time.sleep(self._process_time)
        self._pu.exit()
        _logger.debug("HeatFileProcessing del")

    def setHeatingData(self, hf_stages):
        """!
        Mit dieser Funktion wird das Heizprofil eingelesen.
        Daten werden nur übernommen, wenn kein Heizprogramm läuft.
        Wenn ein Heizprogramm läuft werden nur die Daten der noch nicht ausgeführten Stufen übernommen.
        @param hf_stages: Heizprofil das abgefahren werden soll.
        @return True, wenn Daten übernommen wurden.
        """
        # Im Idle State alles übernehmen
        if self._state == "IDLE":
            self._hfStages = hf_stages
            return True
        # Sonst nur Daten übernehmen, wenn die Programme gleich lang sind.
        elif len(self._hfStages) == len(hf_stages):
            # Daten ab aktiver Stufe +1 bis zum Ende übernehmen
            for i in xrange(self._activeStage+1,len(hf_stages)):
                self._hfStages[i] = hf_stages[i]
            return True
        else:
            return False

    def setUserInteraction(self):
        """!
        Wird in der Haltephase auf eine Benutzerinteraktion gewartet,
        wird das Programm nach Aufruf dieser Funktion fortgesetzt.
        """
        self._userInteraction = True

    def startHeatingProcess(self, stage=0):
        """!
        Wird diese Funktion aufgerufen, startet die Statemachine den Heizprozess.
        @param stage: Stufe, in der das Programm gestartet werden soll.
        """
        if not self._startHeatingProcess and self._state == "IDLE":
            self._last_target_temperature = MIN_TARGET_TEMPERATURE
            self._target_temperature = MIN_TARGET_TEMPERATURE
            self._heat_rate = 0
            self._hold_time = 0
            self._activeStage = stage
            self._startHeatingProcess = True
        else:
            _logger.warning("heat process already running")

    def setPreDryPhase(self, duration=0):
        """!
        Aktiviert die Vortrocknung, wenn eine Zeit > 0 angegeben wird. Mit 0 wird die Vortrocknung deaktiviert.
        @param duration: Vortrocknungszeit in [s].
        """
        if duration >= 0:               # negative Werte abfangen
            self._preDryDuration = duration
        else:
            self._preDryDuration = 0

    def abortHeatingProcess(self):
        """!
        Wird diese Funktion aufgerufen, wird der laufende Heizprozess abgebrochen.
        """
        self._state = "IDLE"
        self._heatingResult = False                         # Abbruch, nicht erfolgreich
        self._pu.disablePowerOutput()                       # Leistungsteil ausschalten
        self._gs1Controller.do_prot_gas('off')
        self._last_target_temperature = MIN_TARGET_TEMPERATURE
        self._target_temperature = MIN_TARGET_TEMPERATURE
        self._heat_rate = 0
        self._hold_time = 0
        self._actuator_set_point = 0
        self._startHeatingProcess = False
        self._estimated_heating_time = 0
        # Je nach Brennkammertemperatur soll die Türe geöffnet oder geschlossen werden.
        if self._analog_input.temp_tc_one_filtered < self._max_temp_door_open + self._set_point_tolerance:
            self._doorController.door_open()
            _logger.debug("opening door")
        else:
            if not self._doorController.is_door_closed():
                self._doorController.door_close()
                _logger.debug("closing door - temperature above safety level")

    def getHeatingProcessTime(self):
        """!
        Gibt Zeit seit dem Start des Heizvorgangs zurück.
        @return Aktuelle Prozesslaufzeit [s].
        """
        if self._state != "IDLE":
            return time.time() - self._heating_start_time
        else:
            return 0

    def getEstimatedProcessTime(self):
        """!
        Gibt die geschätzte Ausführungszeit [s] für das aktuelle Heizprogramm aus.
        @return Geschätzte Restlaufzeit [s].
        """
        return self._estimated_heating_time

    def getSetpointTemperature(self):
        """!
        Gibt Solltemperatur für das aktuelle Heizprogramm aus.
        @return Solltemperatur [K].
        """
        return self._last_target_temperature

    def getCurrentTemperature(self):
        """!
        Gibt die Istemperatur für das aktuelle Heizprogramm aus.
        @return Istemperatur [K].
        """
        current_temperature = self._analog_input.temp_tc_one_filtered
        if (self._state != "IDLE" and
                self._target_temperature - self._set_point_tolerance <= current_temperature <= self._target_temperature + self._set_point_tolerance):
            current_temperature = self._target_temperature

        return current_temperature

    def getPowerUnit(self):
        """!
        Gibt die PowerUnit Instanz zurück.
        @return PowerUnit Instanz.
        """
        return self._pu

    def getHeatingProcessState(self):
        """!
        Diese Funktion gibt eine Liste mit Informationen zum aktuellen Stand des Heizprozesses zurück.
        Der Rückgabewert wird jeweils in den States gesetzt.
        @return Liste mit folgendem Inhalt:\n
                 [0] Geschätzte Restlaufzeit in [min]\n
                 [1] Aktive Stufe: Idle = 0, Vortrocken = 'pre', sonst '1'..\n
                 [2] Heizen='heat', Halten='hold', Kühlen='cool'
        """
        return self._heatingProcessState

    def getHeatingResult(self):
        """!
        Diese Funktion gibt das Ergebnis des Heizprozesses zurück.
        @return True - Heizprozess erfolgreich\n
                False - Heizvorgang nicht erfolgreich.
        """
        return self._heatingResult

    def read(self):
        """!
        Mit dieser Funktion holt der Gs1Controller Meldungen des Heat File Processings an die GUI ab.
        @return Meldungen des Heat File Processings an die GUI. Wenn None, liegt keine Meldung vor
        """
        if self._output_buffer.empty():
            return None
        else:
            rc = self._output_buffer.get()
            self._output_buffer.task_done()
            return rc

    def _do_enter_stage(self, current_temperature):
        """!
        Hier werden Funktionen ausgeführt, die bei Beginn einer Stufe erledigt werden müssen.
        @param current_temperature: Aktuelle Brennkammertemperatur
        """
        # Zieltemperatur, Heizrate und Haltezeit aus aktueller Stufe holen
        self._target_temperature = self._hfStages[self._activeStage]['temp']
        self._heat_rate = self._hfStages[self._activeStage]['rate'] / 60.0      # Durch 60, da in [K/min]
        self._hold_time = self._hfStages[self._activeStage]['hold']

        # Rampengenerator initialisieren
        # Alternative: self._rate_limiter = RateLimiter(self._heat_rate, self._last_target_temperature, RATING_INTERVAL)
        self._rate_limiter = RateLimiter(self._heat_rate, current_temperature)

        # Lineares Abkühlen freigeben, wenn Zieltemperatur kleiner als letzte Zieltemperatur
        self._linear_cooling_enable = True if self._target_temperature < self._last_target_temperature else False

        # Türe zwangsöffnen?
        if ('door' in self._hfStages[self._activeStage]) and self._hfStages[self._activeStage]['door'] == 'force_open':
            self._doorController.door_open()
            _logger.info("force open door")
        # oder beim Heizen automatisch oder auf Befehl schließen
        elif ((self._target_temperature >= self._last_target_temperature) and (not self._doorController.is_door_closed())) or \
             (('door' in self._hfStages[self._activeStage]) and self._hfStages[self._activeStage]['door'] == 'close'):
            self._doorController.door_close()
            _logger.info("closing door")

        # Argon an
        if ('argon' in self._hfStages[self._activeStage]) and self._hfStages[self._activeStage]['argon']:
            _logger.info("argon enabled")
            self._gs1Controller.do_prot_gas('on')
        # Unterdruck an
        if ('vacuum' in self._hfStages[self._activeStage]) and self._hfStages[self._activeStage]['vacuum']:
            _logger.info("vacuum enabled")
            self._gs1Controller.do_vacuum('on')
        # Zwangskühlung an
        if ('forcecooling' in self._hfStages[self._activeStage]) and self._hfStages[self._activeStage]['forcecooling']:
            _logger.info("force cooling enabled")
            self._gs1Controller.do_fc('on')
        # Entlüftung an
        if ('vent' in self._hfStages[self._activeStage]) and self._hfStages[self._activeStage]['vent']:
            _logger.info("ventilation enabled")
            self._gs1Controller.do_vent('on')
        # Lineares Abkühlen an oder aus
        if 'linearcooling' in self._hfStages[self._activeStage]:
            if not self._hfStages[self._activeStage]['linearcooling']:
                self._linear_cooling_enable = False
                _logger.debug("linear cooling disabled")
            else:
                _logger.debug("linear cooling enabled")

    def _do_exit_stage(self, current_temperature):
        """!
        Hier werden Funktionen ausgeführt, die beim Verlassen einer Stufe erledigt werden müssen.
        @param current_temperature: Aktuelle Brennkammertemperatur
        """
        # Türe auf
        if ('door' in self._hfStages[self._activeStage]) and self._hfStages[self._activeStage]['door'] == 'open':
            if current_temperature < self._max_temp_door_open + self._set_point_tolerance:
                self._doorController.door_open()
                _logger.debug("opening door")
            else:
                _logger.debug("Can not open door - temperature above safety level")
        # Argon nach jeder Stufe aus
        self._gs1Controller.do_prot_gas('off')
        # Unterdruck aus
        if ('vacuum' in self._hfStages[self._activeStage]) and not self._hfStages[self._activeStage]['vacuum']:
            _logger.debug("vacuum disabled")
            self._gs1Controller.do_vacuum('off')
        # Zwangskühlung aus
        if ('forcecooling' in self._hfStages[self._activeStage]) and not self._hfStages[self._activeStage]['forcecooling']:
            _logger.debug("force cooling disabled")
            self._gs1Controller.do_fc('off')
        # Entlüftung aus
        if ('vent' in self._hfStages[self._activeStage]) and not self._hfStages[self._activeStage]['vent']:
            _logger.debug("ventilation disabled")
            self._gs1Controller.do_vent('off')

    def _linearCooling(self, current_temperature, pid):
        """!
        Funktion für lineares Abkühlen. Öffnet die Türe, wenn die Brennkammer nicht schnell genug abkühlt.
        @param current_temperature: Aktuelle Brennkammertemperatur
        @param pid: Werteliste aus PID Regler
        """
        if not self._linear_cooling_timer_set:
            self._next_linear_cooling = time.time() + self._door_remain_time_cooling
            self._linear_cooling_timer_set = True

        if time.time() >= self._next_linear_cooling:
            if (int(pid[0]) <= self._k_p_lin_cooling) and (current_temperature < self._door_max_temp_cooling):
                if self._doorController.actual_percent_door < self._linear_cooling_max_value:
                    self._doorController.door_open(self._door_step_value)
                    _logger.debug('linear cooling: opened door by {} percent'.format(self._door_step_value))
            self._linear_cooling_timer_set = False

    def _temperature_monitoring(self, current_temperature):
        """!
        Funktion zur Überwachung der Prozesstemperatur:\n
          - Trägt bei Abweichungen entsprechende Fehler ein.\n
          - Fehlerreaktion laut Fehlerliste.
        @param current_temperature: Aktuelle Brennkammertemperatur
        """

        # Überwachung: Endtemperatur nicht erreichbar in den Heizphasen
        if ((self._state == "GET_PREDRY_TEMPERATURE" or self._state == "GET_TEMPERATURE") and
           (self._target_temperature > self._last_target_temperature)):
            delta_temp = current_temperature - self._last_F701_temp     # Temperaturdifferenz zur letzten Messung berechnen
            if delta_temp < self._min_heating_gradient:                 # Bei einer Abweichung von weniger als _min_heating_gradient
                if self._target_temp_timeout < time.time():             # auf Timeout warten
                    _logger.debug("F701: {}, {}, {}, {}, {}, {}, {}".format(self._state, self._target_temperature,
                                                                            self._last_target_temperature,
                                                                            current_temperature, self._last_F701_temp,
                                                                            self._target_temp_timeout, time.time()))
                    # TODO Analyse: self._heatingResult = False                         # Heizprozess nicht erfolgreich
                    # TODO Analyse: self._error_handler.set_error("F701")               # und Fehler eintragen.
            else:                                                       # Bei einer Abweichung von mehr als 1K
                self._target_temp_timeout = time.time() + self._f701_timeout  # Timeout neu starten
                self._last_F701_temp = current_temperature              # und Temperatur merken

        # Überwachung: Temperaturfehler und -abweichung in den Haltephasen
        if ((self._state == "HOLD_PREDRY_TEMPERATURE") or
           (self._state == "HOLD_TEMPERATURE")):
            # Ausnahme: Zwangsöffnen bei Superspeed
            if not('door' in self._hfStages[self._activeStage] and self._hfStages[self._activeStage]['door'] == 'force_open'):
                # Abweichung von mehr als +/-30K
                if not (self._target_temperature - F70_TEMP_DEVIATION) <= current_temperature <= (self._target_temperature + F70_TEMP_DEVIATION):
                    self._heatingResult = False                             # Heizprozess nicht erfolgreich
                    self.abortHeatingProcess()
                    self._error_handler.set_error("F70")                    # Fehler F70 eintragen
                # Abweichung von +/-10K bis +/-30K
                elif (not (self._target_temperature - F700_TEMP_DEVIATION) < current_temperature < (self._target_temperature + F700_TEMP_DEVIATION) and
                     (self._state == "HOLD_TEMPERATURE")):                      # Auf Wunsch von Herrn Friedrich nicht mehr bei Vortrocknen.
                        self._heatingResult = True                             # Heizprozess trotz Fehler F700 erfolgreich
                        self._error_handler.set_error("F700")                   # Fehler F700 eintragen

    def _calculateEstimatedProcessTime(self, current_stage, current_temperature, current_hold_time):
        """!
        Funktion berechnet die geschätzte verbleibende Prozessdauer in [s].
        @param current_stage: Aktuelle Stufe
        @param current_temperature: Aktuelle Brennkammertemperatur
        @param current_hold_time: Aktuelle Haltezeit
        @return Geschätzte Restlaufzeit [s]
        """
        estimatedprocesstime = 0
        ramp_time = 0
        last_target_temperature = current_temperature
        # Zieltemperatur, Heizrate und Haltezeit aus aktueller Stufe holen
        target_temperature = self._hfStages[current_stage]['temp']
        heat_rate = self._hfStages[current_stage]['rate'] / 60.0        # Durch 60, da in [K/min]
        hold_time = self._hfStages[current_stage]['hold']

        # Abhängig von aktueller Phase eine andere Berechnung durchführen
        # Heizphase Vortrocknen
        if (self._state == "PREPARE_PREDRY_PHASE" or
            self._state == "GET_PREDRY_TEMPERATURE"):
            hold_time += self._preDryDuration                           # Haltezeit
            # Heiz-/Abkühlzeit
            ramp_time = (self._pre_dry_temperature - last_target_temperature) / self._pre_dry_heatingrate
        # Haltephase Vortrocknen
        elif self._state == "HOLD_PREDRY_TEMPERATURE":
            hold_time += (self._preDryDuration - current_hold_time)     # verbleibende Haltezeit berechnen
        # Heiz-/Kühlphasen
        elif self._state == "GET_TEMPERATURE":
            # Heiz-/Abkühlzeit berechnen
            ramp_time = (target_temperature - last_target_temperature) / heat_rate
            current_stage += 1                                          # nächste Stufe
        # Haltephasen
        elif self._state == "HOLD_TEMPERATURE":
            hold_time -= current_hold_time                              # verbleibende Haltezeit berechnen
            current_stage += 1                                          # nächste Stufe

        # Heiz-/Abkühlzeit zu Gesamtzeit addieren, Betrag bilden, da beim Kühlen negativ.
        estimatedprocesstime += abs(ramp_time)
        # Haltezeit zu Gesamtzeit addieren
        estimatedprocesstime += hold_time

        # Berechnung für alle verbleibenden Stufen
        for stageIdx in xrange(current_stage, len(self._hfStages)):
            # nächste Stufe, alte Zieltemperatur sichern
            last_target_temperature = target_temperature

            # neue Stufe lesen
            target_temperature = self._hfStages[stageIdx]['temp']
            heat_rate = self._hfStages[stageIdx]['rate'] / 60.0         # Durch 60, da in [K/min]
            hold_time = self._hfStages[stageIdx]['hold']

            # TODO: Was machen bei Benutzerinteraktion?
            if hold_time < 0:
                hold_time = 0

            # Heiz-/Abkühlzeit
            ramp_time = (target_temperature - last_target_temperature) / heat_rate

            # Heiz-/Abkühlzeit zu Gesamtzeit addieren, Betrag bilden, da beim Kühlen negativ.
            estimatedprocesstime += abs(ramp_time)
            # Haltezeit zu Gesamtzeit addieren
            estimatedprocesstime += hold_time

        return estimatedprocesstime

    def _process(self):
        """!
        Heat File Processing Thread.
        """
        _logger.debug("{} process started".format(self._name))

        t_hold_time_start = time.time()
        delta_t_hold_time = 0
        pid = [0, 0]

        self._next_linear_cooling = time.time() + self._door_remain_time_cooling
        self._linear_cooling_timer_set = False

        _1s_cycle_time = time.time()

        wait_time = self._process_time

        while not self._stopped.wait(wait_time):
            process_start_time = time.time()
            current_temperature = self._analog_input.temp_tc_one_filtered

            #############################################
            # IDLE State                                #
            #############################################
            if self._state == "IDLE":
                self._heatingProcessState = ['0', '0', 'hold']
                if self._startHeatingProcess:
                    _1s_cycle_time = time.time()
                    self._startHeatingProcess = False
                    self._heatingResult = True                  # Annahme Heizvorgang erfolgreich
                    self._heating_start_time = time.time()
                    self._target_temp_timeout = time.time() + self._f701_timeout    # F701 Timeout neu starten
                    self._last_F701_temp = current_temperature                      # und Temperatur merken
                    if self._hfStages:
                        self._error_handler.reset_error("F8000")
                        self._error_handler.reset_error("F80")
                        self._pid.clear()                       # PID Regler Reset
                        self._pu.enablePowerOutput()            # Leistungsteil aktivieren
                        if self._preDryDuration > 0:            # Soll eine Vortrockung stattfinden?
                            _logger.debug("preparing pre dry phase")
                            self._heatingProcessState = [str(int(ceil(self._estimated_heating_time / 60))), 'pre',
                                                         'heat']
                            if not self._doorController.is_door_closed():
                                self._doorController.door_close()
                                _logger.debug("closing door for predry preparation")
                            self._target_temperature = self._pre_dry_temperature
                            # Rampengenerator initialisieren
                            self._rate_limiter = RateLimiter(self._pre_dry_heatingrate, self._last_target_temperature)
                            self._hold_time = self._preDryDuration
                            self._state = "PREPARE_PREDRY_PHASE"
                        else:                                   # Oder direkt das Heizprofil gestartet werden?
                            _logger.debug("heating process started")
                            self._state = "ENTER_STAGE"

                    else:
                        self.abortHeatingProcess()
            else:
                #############################################
                # Rampengenerator, PID-Regler,              #
                # Stromregelung und Leistungsteil nur       #
                # aktiv, wenn nicht IDLE                    #
                #############################################
                if _1s_cycle_time < time.time():    # 1s Timer abgelaufen
                    _1s_cycle_time += 1             # Neuen Timer setzen
                    # Temperaturbegrenzung, wenn Türe offen ist
                    max_allowed_temperature = self._target_temperature
                    # Maximal erlaubt ist die maximale Öffnung für lineares Kühlen + eine Schrittweite
                    if self._doorController.actual_percent_door > self._linear_cooling_max_value + self._door_step_value:
                        if self._target_temperature > self._max_temp_door_open:
                            # Ausnahme: Zwangsöffnung bei z.B. Superspeed Verfahren oder Vortrocken-Phase
                            if not(('door' in self._hfStages[self._activeStage] and self._hfStages[self._activeStage]['door'] == 'force_open') or
                               self._state == "GET_PREDRY_TEMPERATURE" or self._state == "HOLD_PREDRY_TEMPERATURE"):
                                max_allowed_temperature = self._max_temp_door_open
                                _logger.debug('door open, limit max temperature: {} K'.format(self._max_temp_door_open))

                    # Rampengenerator jede Sekunde ausführen
                    self._last_target_temperature = self._rate_limiter.rate(max_allowed_temperature)
                    # PID Regler jede Sekunde ausführen
                    # Reglerparameter kp, ki und kd in Abhängigkeit der aktuellen Brennkammertemperatur wechseln.
                    self._pid.updatePIDParameters(current_temperature)
                    # Reglerparameter kp, ki und kd in Abhängigkeit der gewünschten Solltemperatur wechseln.
                    # self._pid.updatePIDParameters(self._target_temperature)
                    # Reglerparameter kp, ki und kd in Abhängigkeit der aktuellen Brennkammertemperatur wechseln.
                    pid = self._pid.update(self._last_target_temperature, current_temperature)
                    if ENABLE_DEBUG_LOG:
                        _logger.debug("PID output: {}".format(pid))

                    # Nur in den genannten States die Restlaufzeit berechnen und Temperatur überwachen
                    if ((self._state == "GET_PREDRY_TEMPERATURE") or
                       (self._state == "HOLD_PREDRY_TEMPERATURE") or
                       (self._state == "PREPARE_PREDRY_PHASE") or
                       (self._state == "GET_TEMPERATURE") or
                       (self._state == "HOLD_TEMPERATURE")):
                        # Geschätzte Restlaufzeit berechnen
                        self._estimated_heating_time = self._calculateEstimatedProcessTime(self._activeStage, current_temperature, delta_t_hold_time)
                        # Temperaturüberwachung
                        self._temperature_monitoring(current_temperature)

                # Adaptive Stromregelung begrenzt bei Bedarf den Sollwert aus PID-Regler
                actuator_set_point = self._asr(current_temperature, self._pu.getCurrentEff, self._pu.getPwrActualValue, int(pid[1]))
                # Leistungseinheit mit begrenztem Wert ansteuern
                self._pu.setPowerOutput(actuator_set_point)

                # Wenn die Türisolierung fehlt, dann SOFORT abbrechen!
                if self._doorController.is_door_insulation_missing():
                    _logger.debug("abort heat process, door insulation is missing!")
                    self.abortHeatingProcess()

                if self._error_handler.is_error_active("F80"):
                    _logger.debug("abort heat process, argon pressure error!")
                    self._output_buffer.put("stargon==fail")
                    self.abortHeatingProcess()

                self._startHeatingProcess = False

            #############################################
            # Türe auf 50% einstellen                   #
            #############################################
            if self._state == "PREPARE_PREDRY_PHASE":
                # Aktuelle Temperatur muss in definiertem Toleranzbereich sein, damit in die Haltephase gewechselt wird
                # In der Vortrocknung nur größer als Solltemperatur
                self._heatingProcessState = [str(int(ceil(self._estimated_heating_time/60))), 'pre', 'heat']
                _logger.debug("pre dry phase started")
                if self._doorController.is_door_closed():
                    if self._analog_input.temp_tc_one_filtered < self._max_temp_door_open:
                        self._doorController.door_open(50)      # 28.01.2019: Anforderung von Sirona: Türe beim Vortrocken zu 50% öffnen
                        _logger.debug("set door to 50%")
                        self._state = "GET_PREDRY_TEMPERATURE"
                    else:
                        _logger.debug("abort pre dry phase and heat process, temperature to high!")
                        self.abortHeatingProcess()

            #############################################
            # Heizen auf Vortrocknungstemperatur        #
            #############################################
            if self._state == "GET_PREDRY_TEMPERATURE":
                # Aktuelle Temperatur muss in definiertem Toleranzbereich sein, damit in die Haltephase gewechselt wird
                # In der Vortrocknung nur größer als Solltemperatur
                self._heatingProcessState = [str(int(ceil(self._estimated_heating_time/60))), 'pre', 'heat']
                if self._target_temperature - self._set_point_tolerance <= current_temperature:
                    t_hold_time_start = time.time()
                    _logger.debug("temperature {}K reached, pre dry hold phase started".format(self._target_temperature))
                    self._state = "HOLD_PREDRY_TEMPERATURE"

            #############################################
            # Haltephase Vortrocknung                   #
            #############################################
            if self._state == "HOLD_PREDRY_TEMPERATURE":
                self._heatingProcessState = [str(int(ceil(self._estimated_heating_time/60))), 'pre', 'hold']
                delta_t_hold_time = time.time() - t_hold_time_start
                if delta_t_hold_time >= self._hold_time:
                    self._state = "ENTER_STAGE"
                    _logger.debug("pre dry phase done, heating process started")

            #############################################
            # Einleitung einer Stufe                    #
            #############################################
            if self._state == "ENTER_STAGE":
                _logger.debug("entered stage {}".format(self._activeStage))
                # ENTER_STAGE Funktionen ausführen
                self._do_enter_stage(current_temperature)

                # Prüfen, ob die Stufe wegen bereits erreichter Temperatur übersprungen werden kann
                # TODO: Was machen, wenn in der zu überspringenden Stufe eine Benutzerinteraktion erwartet wird?
                if (self._target_temperature > self._last_target_temperature) and (self._target_temperature < current_temperature):
                    _logger.debug("Temperature {}K already reached, skipping stage {}".format(self._target_temperature, self._activeStage))
                    self._last_target_temperature = self._target_temperature    # Zieltemperatur der übersprungenen Stufe übernehmen
                    # Rampengenerator mit neuer Starttemperatur initialisieren
                    self._rate_limiter = RateLimiter(self._heat_rate, self._last_target_temperature)
                    self._state = "EXIT_STAGE"
                else:
                    _logger.debug("heating up/cooling down to {}K".format(self._target_temperature))
                    self._state = "GET_TEMPERATURE"

            #############################################
            # Heizen/Kühlen auf Solltemperatur          #
            #############################################
            if self._state == "GET_TEMPERATURE":
                self._heatingProcessState = [str(int(ceil(self._estimated_heating_time/60))), str(self._activeStage+1), 'heat']
                # Aktuelle Temperatur muss in definiertem Toleranzbereich sein, damit in die Haltephase gewechselt wird
                if self._target_temperature - self._set_point_tolerance <= current_temperature <= self._target_temperature + self._set_point_tolerance:
                    t_hold_time_start = time.time()
                    _logger.debug("temperature {}K reached, hold phase started".format(self._target_temperature))

                    # Beim eintritt in die Haltephase die Türe wieder zu machen.
                    # So wird verhindert, dass die Sicehrheitstemperaturbegrenzung aktiv werden kann.
                    if (not self._doorController.is_door_closed()):
                        self._doorController.door_close()
                        _logger.info("closing door")

                    self._state = "HOLD_TEMPERATURE"
                    if self._hold_time < 0:
                        self._userInteraction = False
                        self._output_buffer.put("waitforuserinteraction")
                else:
                    # Lineares Abkühlen
                    if self._linear_cooling and self._linear_cooling_enable:
                        self._linearCooling(current_temperature, pid)

            #############################################
            # Haltephase Solltemperatur                 #
            #############################################
            if self._state == "HOLD_TEMPERATURE":
                if self._hold_time < 0:
                    self._heatingProcessState = [str(int(ceil(self._estimated_heating_time/60))), str(self._activeStage+1), 'user']
                    if self._userInteraction:
                        _logger.debug("temperature hold phase quit due to user interaction")
                        self._state = "EXIT_STAGE"
                else:
                    self._heatingProcessState = [str(int(ceil(self._estimated_heating_time/60))), str(self._activeStage+1), 'hold']
                    delta_t_hold_time = time.time() - t_hold_time_start
                    if delta_t_hold_time >= self._hold_time:
                        self._state = "EXIT_STAGE"
                        _logger.debug("temperature hold phase quit due to timeout")

            #############################################
            # Beenden einer Stufe                       #
            #############################################
            if self._state == "EXIT_STAGE":
                # EXIT_STAGE Funktionen ausführen
                self._do_exit_stage(current_temperature)
                _logger.debug("leaving stage {}".format(self._activeStage))

                # Nächste Stufe vorhanden oder Programm beenden?
                self._activeStage = self._activeStage + 1
                if self._activeStage < len(self._hfStages):
                    self._state = "ENTER_STAGE"
                else:
                    self._pu.disablePowerOutput()           # Leistungsteil ausschalten
                    self._gs1Controller.do_prot_gas('off')
                    self._startHeatingProcess = False
                    _logger.debug("heating process done")
                    if (current_temperature <= self._max_temp_door_open + self._set_point_tolerance or
                       ('door' in self._hfStages[self._activeStage-1] and self._hfStages[self._activeStage-1]['door'] == 'force_open')):
                        self._doorController.door_open()
                        _logger.debug("opening door")
                    else:
                        _logger.debug("Can not open door - temperature above safety level")
                    self._state = "IDLE"

            # Ständige Überwachung von Übertemperatur
            if current_temperature >= self._limit_temperature_cc:
                self._heatingResult = False                                 # Heizprozess nicht erfolgreich
                self._error_handler.set_error("F71")                        # maximale Brennkammertemperatur überschritten

            # Auswertung, wie lange die Ausführung gedauert hat.
            _now_ = time.time()
            delta_process_time = _now_ - process_start_time
            wait_time = self._process_time - delta_process_time
            if ENABLE_DEBUG_LOG:
                _logger.debug("time: {}, delta_process_time: {}, wait_time: {}".format(_now_, delta_process_time, wait_time))
            if wait_time <= 0:
                wait_time = 0
                if ENABLE_DEBUG_LOG:
                    _logger.debug("no wait time:{}".format(delta_process_time))
